from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestHorizontalLineSegment(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hls = HorizontalLineSegment(length=10)

    def test_get_relative_x2(self):
        actual = self.hls.get_relative_x2()
        expected = self.hls.length
        self.assertEqual(expected, actual)

    def test_start_mark_line_relative_x(self):
        actual = self.hls.start_mark_line.relative_x
        expected = self.hls.relative_x
        self.assertEqual(expected, actual)

    def test_start_mark_line_relative_y(self):
        actual = self.hls.start_mark_line.relative_y
        expected = -1.5
        self.assertEqual(expected, actual)

    def test_end_mark_line_relative_x(self):
        actual = self.hls.end_mark_line.relative_x
        expected = self.hls.relative_x + self.hls.length
        self.assertEqual(expected, actual)

    def test_end_mark_line_relative_y(self):
        actual = self.hls.end_mark_line.relative_y
        expected = -1.5
        self.assertEqual(expected, actual)

    def test_straight_line_top_margin(self):
        actual = self.hls.straight_line.top_margin
        expected = 0
        self.assertEqual(expected, actual)

    def test_start_mark_line_top_margin(self):
        actual = self.hls.start_mark_line.top_margin
        expected = 0
        self.assertEqual(expected, actual)

    def test_get_height(self):
        actual = self.hls.get_height()
        expected = 3
        self.assertEqual(expected, actual)

    def test_draw(self):
        with self.file_path(parent_path=path, name='draw', extension='pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            self.hls.end_mark_line.show = True
            self.hls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_list(self):
        segments = [HorizontalLineSegment(length) for length in range(10, 30, 5)]
        segments[-1].end_mark_line.show = True
        with self.file_path(parent_path=path, name='draw_list', extension='pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            for segment in segments:
                segment.draw(self.pdf)
                self.pdf.translate(segment.get_width(), 0)

            self.pdf.write(pdf_path)

    def test_draw_with_top_margin(self):
        self.hls.top_margin = 15

        with self.file_path(parent_path=path, name='draw_with_top_margin', extension='pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)

            self.hls.end_mark_line.show = True
            self.hls.draw(self.pdf)

            self.pdf.write(pdf_path)



import copy
from itertools import chain

from quicktions import Fraction

from musurgia.arithmeticprogression import ArithmeticProgression, DAndSError
from musurgia.basic_functions import dToX


class ValueGeneratorException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class GeneratorHasNoNextError(ValueGeneratorException):
    def __init__(self, generator, *args):
        msg = 'generator {} has no __next__'.format(generator)
        super().__init__(msg, *args)


class GeneratorNotIterableError(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class XOutOfRange(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class NoDurationError(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class ArithmeticConflictError(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class PositionError(ValueGeneratorException):
    def __init__(self, position, duration, *args):
        msg = 'position {} must be smaller than duration {}'.format(position, duration)
        super().__init__(msg, *args)


class CallConflict(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class ChildrenValueModeConflict(ValueGeneratorException):
    def __init__(self, children_value_modes, *args):
        msg = 'children have more than one value_mode: {}'.format(children_value_modes)
        super().__init__(msg, *args)


class ValueGeneratorTypeConflict(ValueGeneratorException):
    def __init__(self, *args):
        super().__init__(*args)


class ValueGenerator(object):
    def __init__(self, generator=None, value_mode=None, duration=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._generator = None
        self._duration = None
        self._position = 0
        self._children = None
        self._children_iterator = None

        self.generator = generator
        self.value_mode = value_mode
        self.duration = duration

    @property
    def generator(self):
        return self._generator

    @generator.setter
    def generator(self, val):
        if val and not callable(val) and not hasattr(val, '__next__'):
            raise TypeError('generator {} must have  __call__ or __next__ attribute'.format(val))
        self._generator = val

    def _set_generator_duration(self):
        if isinstance(self.generator, ArithmeticProgression) and self.value_mode in ['duration']:
            try:
                self.generator.s = self.duration
            except DAndSError:
                if self.generator.s != self.duration:
                    raise ArithmeticConflictError()
        elif hasattr(self.generator, 'quarter_duration'):
            self.generator.quarter_duration = self.duration
        elif hasattr(self.generator, 'duration'):
            self.generator.duration = self.duration
        else:
            pass

    @property
    def value_mode(self):
        if self.children:
            if self._value_mode is None:
                children_value_modes = list(
                    dict.fromkeys([child.value_mode for child in self.children if child.value_mode is not None])
                )
                if not children_value_modes:
                    return None
                elif len(children_value_modes) > 1:
                    raise ChildrenValueModeConflict(children_value_modes)
                else:
                    return children_value_modes[0]
            else:
                for child in self.children:
                    child.value_mode = self._value_mode

        return self._value_mode

    @value_mode.setter
    def value_mode(self, val):
        permitted = [None, 'duration', 'midi', 'chord']
        if val not in permitted:
            raise ValueError('value_mode.value {} must be in {}'.format(val, permitted))
        self._value_mode = val

    @property
    def duration(self):
        if self.children:
            children_durations = [child.duration for child in self.children]
            if None in children_durations:
                return None
            else:
                return sum(children_durations)
        return self._duration

    @duration.setter
    def duration(self, val):
        if val is not None:
            if self.children:
                raise ValueGeneratorException('parent\'s duration cannot be set')
                # children_duration = [fractal_tree.duration for fractal_tree in self.children]
                # if None in children_duration:
                #     raise ValueGeneratorException(
                #         'ValueGenerator: parent\'s duration cannot be set if not all children have a duration.')
                # factor = Fraction(val, sum(children_duration))
                # for fractal_tree in self.children:
                #     fractal_tree.duration *= factor
        self._duration = val
        self._set_generator_duration()

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        if val < 0:
            raise ValueError(float(val))
        self._position = val

    @property
    def children(self):
        return self._children

    def add_child(self, child):
        if not isinstance(child, ValueGenerator):
            raise TypeError('fractal_tree must be of type ValueGenerator not{}'.format(type(child)))
        if self._children is None:
            self._children = []
        if self._children_iterator is None:
            self._children_iterator = iter([])
        child.parent_group = self
        self._children.append(child)
        self._children_iterator = chain(self._children_iterator, child)

    def _check_position(self):
        if not self.duration:
            raise NoDurationError()
        if not self.position < self.duration:
            raise PositionError(self.position, self.duration)

    def _change_position_in_duration(self, value):
        if self.value_mode == 'duration':
            self.position += value
        elif self.value_mode == 'chord':
            self.position += value.quarter_duration

    def __next__(self):
        try:
            return self.__call__(self.position)
        except (ValueError, PositionError):
            raise StopIteration()

    def __call__(self, x):
        self.position = x
        self._check_position()
        output = None
        if self.children:
            durations = [vg.duration for vg in self.children]
            duration_limits = dToX(durations)
            for i in range(len(duration_limits) - 1):
                if duration_limits[i] <= x < duration_limits[i + 1]:
                    output = self.children[i](x - duration_limits[i])
                    break
        elif callable(self.generator):
            output = self.generator.__call__(x)
        else:
            # if isinstance(self.generator, ArithmeticProgression):
            #     raise CallConflict('Calling Arithmetic Progression is not allowed.')
            output = self.generator.__next__()
        self._change_position_in_duration(output)
        return output

    def __iter__(self):
        return self

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(value_mode=self.value_mode,
                                duration=self.duration)
        if self.generator is not None:
            copied.generator = copy.deepcopy(self.generator)
        if self.children is not None:
            for child in self.children:
                copied.add_child(child.__deepcopy__())
        return copied

# class ValueGeneratorGroup(object):
#     def __init__(self, *value_generators, **kwargs):
#         super().__init__(**kwargs)
#         self._value_generators = None
#         self._value_generators_iterator = None
#         self._child_type = None
#         self._current_value_generator = None
#         self.value_generators = value_generators
#
#     @property
#     def value_generators(self):
#         return self._value_generators
#
#     @value_generators.setter
#     def value_generators(self, values):
#         try:
#             values = list(values)
#         except TypeError:
#             values = [values]
#
#         for value in values:
#             self.add_value_generator(value)
#
#     @property
#     def duration(self):
#         try:
#             return sum([vg.duration for vg in self.value_generators])
#         except AttributeError:
#             return 0
#
#     def add_value_generator(self, value_generator):
#         if not isinstance(value_generator, ValueGenerator):
#             raise TypeError('value_generators must be of type ValueGenerator not{}'.format(type(value_generator)))
#         if self._value_generators is None:
#             self._value_generators = []
#         if self._value_generators_iterator is None:
#             self._value_generators_iterator = iter([])
#         value_generator.parent_group = self
#         self._value_generators.append(value_generator)
#         self._value_generators_iterator = chain(self._value_generators_iterator, value_generator)
#
#     def __next__(self):
#         return self._value_generators_iterator.__next__()
#
#     def __iter__(self):
#         return self
#
#     def __call__(self, x):
#         durations = [vg.duration for vg in self.value_generators]
#         duration_limits = dToX(durations)
#         for i in range(len(duration_limits) - 1):
#             if duration_limits[i] <= x < duration_limits[i + 1]:
#                 return self.value_generators[i](x - duration_limits[i])



from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment, HorizontalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestMarkLineLabels(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.ls = HorizontalLineSegment(length=20)

    def test_draw_above(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above')
        ml.add_text_label('second text label above')
        ml.add_text_label('third  text label above')
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, 'draw_above', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')

            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_below(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label below', placement='below')
        ml.add_text_label('second text label below', placement='below')
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, 'draw_below', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_left(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label left', placement='left')
        ml.add_text_label('second text label left left left', placement='left')
        ml.add_text_label('third text label left left left', placement='left')
        # for text_label in ml.left_text_labels:
        # text_label.right_margin = 2
        # text_label.top_margin = -2
        ml.left_text_labels[1].font.size = 8
        self.ls.relative_x = 40
        self.ls.relative_y = 10
        with self.file_path(path, 'draw_left', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_different_sizes(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above', font_size=7, bottom_margin=2)
        ml.add_text_label('second text label above', font_size=8, bottom_margin=4)
        ml.add_text_label('third text label above', font_size=9)
        with self.file_path(path, 'different_sizes', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_font_size_8(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above', font_size=8, bottom_margin=2)
        ml.add_text_label('second text label above', font_size=8, bottom_margin=4)
        ml.add_text_label('third text label above', font_size=8)
        with self.file_path(path, 'font_size_8', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_below_with_different_mark_line_lengths(self):
        hsl = HorizontalSegmentedLine(lengths=[10, 15, 20])
        hsl.segments[0].start_mark_line.length = 6
        hsl.segments[0].start_mark_line.add_label(1, placement='below', font_size=8)
        hsl.segments[1].start_mark_line.add_label(2, placement='below', font_size=8)
        with self.file_path(path, 'below_with_different_mark_line_lengths', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            hsl.draw(self.pdf)
            self.pdf.write(pdf_path)



from math import floor, ceil
from quicktions import Fraction


def _find_nearest_quantized_value(quantized_values, values):
    output = []
    for value in values:
        nearest_quantized = min(enumerate(quantized_values), key=lambda x: abs(x[1] - value))[1]
        delta = nearest_quantized - value
        output.append((nearest_quantized, delta))
    return output


def _find_quantized_locations(positions, grid_size):
    def get_quantized(val, key):
        factor = 1 / grid_size

        if key == "min":
            output = ceil(val * factor) / factor
        elif key == "max":
            output = floor(val * factor) / factor
        else:
            raise ValueError()
        return output

    min_location = Fraction(get_quantized(min(positions), "min"))
    output = [min_location]
    max_location = Fraction(get_quantized(max(positions), "max"))

    while output[-1] < max_location:
        output.append(output[-1] + Fraction(grid_size))

    return output


def get_quantized_positions(positions, grid_size):
    def _get_quantized_locations():
        return _find_quantized_locations(positions, grid_size)

    quantized_positions = [f[0] for f in _find_nearest_quantized_value(_get_quantized_locations(), positions)]

    return quantized_positions


def get_quantized_values(vals, grid_size):
    # print("vals", vals)
    def _get_positions():
        positions = [0]
        for val in vals:
            positions.append(positions[-1] + val)
        return positions

    positions = _get_positions()
    quantized_positions = get_quantized_positions(positions, grid_size)
    quantized_vals = []
    for index, val in enumerate(vals):
        quantized_val = Fraction(quantized_positions[index + 1] - quantized_positions[index]).limit_denominator(100)
        quantized_vals.append(quantized_val)

    return quantized_vals



from abc import ABC, abstractmethod

from musurgia.pdf.margined import MarginNotSettableError
from musurgia.pdf.positioned import RelativePositionNotSettableError


class _GetSlavePositionMixIn(ABC):
    @abstractmethod
    def get_slave_position(self, slave, position):
        pass


class _GetSlaveMarginMixIn(ABC):
    @abstractmethod
    def get_slave_margin(self, slave, margin):
        pass


class _SetSlavePositionMixIn:
    @property
    def relative_x(self):
        return self.master.get_slave_position(slave=self, position='x')

    @relative_x.setter
    def relative_x(self, val):
        if val is not None:
            raise RelativePositionNotSettableError()

    @property
    def relative_y(self):
        return self.master.get_slave_position(slave=self, position='y')

    @relative_y.setter
    def relative_y(self, val):
        if val is not None:
            raise RelativePositionNotSettableError()


class _SetSlaveMarginMixIn:
    @property
    def left_margin(self):
        return self.master.get_slave_margin(slave=self, margin='left')

    @left_margin.setter
    def left_margin(self, val):
        if val is not None:
            raise MarginNotSettableError()

    @property
    def top_margin(self):
        return self.master.get_slave_margin(slave=self, margin='top')

    @top_margin.setter
    def top_margin(self, val):
        if val is not None:
            raise MarginNotSettableError()

    @property
    def right_margin(self):
        return self.master.get_slave_margin(slave=self, margin='right')

    @right_margin.setter
    def right_margin(self, val):
        if val is not None:
            raise MarginNotSettableError()

    @property
    def bottom_margin(self):
        return self.master.get_slave_margin(slave=self, margin='bottom')

    @bottom_margin.setter
    def bottom_margin(self, val):
        if val is not None:
            raise MarginNotSettableError()


class _Named:
    def __init__(self, name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = None
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        if val and not isinstance(val, str):
            raise TypeError(f"name.value must be of type str not{type(val)}")
        self._name = val


class PositionMaster(_GetSlavePositionMixIn):
    pass


class MarginMaster(_GetSlaveMarginMixIn):
    pass


class Master(_GetSlavePositionMixIn, _GetSlaveMarginMixIn):
    pass


class PositionSlave(_Named, _SetSlavePositionMixIn):
    def __init__(self, master=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._master = None
        self.master = master

    @property
    def master(self):
        return self._master

    @master.setter
    def master(self, val):
        if val and not isinstance(val, PositionMaster):
            raise TypeError(f"master.value must be of type {PositionMaster} not {type(val)}")
        self._master = val


class MarginSlave(_Named, _SetSlaveMarginMixIn):
    def __init__(self, master, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._master = None
        self.master = master

    @property
    def master(self):
        return self._master

    @master.setter
    def master(self, val):
        if val and not isinstance(val, MarginMaster):
            raise TypeError(f"master.value must be of type {MarginMaster} not {type(val)}")
        self._master = val


class Slave(_Named, _SetSlavePositionMixIn, _SetSlaveMarginMixIn):
    def __init__(self, master, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._master = None
        self.master = master

    @property
    def master(self):
        return self._master

    @master.setter
    def master(self, val):
        if val and not isinstance(val, Master):
            raise TypeError(f"master.value must be of type {Master} not {type(val)}")
        self._master = val



from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusicsquare import SquareGroup, Square


class Test(TestCase):
    def setUp(self) -> None:
        self.square_group = SquareGroup(
            Square(duration=10, name='red', tree_permutation_order=(3, 1, 2), proportions=(1, 2, 3)),
            Square(duration=10, name='blue', tree_permutation_order=(3, 1, 2), proportions=(1, 2, 3)),
            Square(duration=10, name='green', tree_permutation_order=(3, 1, 2), proportions=(1, 2, 3)),
        )

    def test_1(self):
        expected = self.square_group.squares[0].rows + self.square_group.squares[1].rows + self.square_group.squares[2].rows
        result = self.square_group.get_all_rows()
        self.assertEqual(expected, result)



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2))
        self.fm.duration = 10
        self.fm.tempo = 60
        self.fm.midi_generator.midi_range = [60, 67]
        self.fm.add_layer()
        self.score = TreeScoreTimewise()
        self.score.set_time_signatures(quarter_durations=[self.fm.quarter_duration])

    def test_1(self):

        sp = self.fm.get_children()[0].split(1, 1, 1, 1, 1)
        sp[0].chord.add_dynamics('pp')
        sp[0].chord.add_slur('start')
        sp[-1].chord.add_slur('stop')
        sf = self.fm.get_simple_format()
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1)

        xml_path = path + '_test_1.xml'
        self.score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_2(self):
        def add_repetition(node):
            node.chord.add_articulation('tenuto')
            duration = node.chord.quarter_duration
            lengths = int(duration) * [1]
            if len(lengths) > 1:
                diff = duration - int(duration)
                if diff > 0:
                    lengths[-1] += diff
                sp = node.split(lengths)
                sp[0].chord.add_slur('start')
                sp[-1].chord.add_slur('stop')

        for child in self.fm.get_children():
            add_repetition(child)

        sf = self.fm.get_simple_format()
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1)

        xml_path = path + '_test_2.xml'
        self.score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)



from unittest import TestCase

from musurgia.timing import Clock, Timing


class Test(TestCase):
    def test_1(self):
        seconds = 180
        self.assertEqual(Timing.get_clock(seconds), '''0:3':00"''')

    def test_2(self):
        t = Timing(quarter_duration=1, duration=0.65)
        result = 92.3076923076923
        self.assertEqual(result, t.tempo)

    def test_3(self):
        clock = Clock()
        clock.duration = 37
        result = (0, 0, 37)
        self.assertEqual(result, clock.clock)

    def test_4(self):
        clock1 = Clock(clock=(0, 2, 31))
        clock2 = Clock(clock=(0, 3, 11))
        clock3 = clock2.subtract(clock1)
        result = (0, 0, 40)
        self.assertEqual(result, clock3.clock)



from pathlib import Path

from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


def make_ft():
    ft = FractalTree(value=20)
    ft.add_layer()
    ft.get_children()[0].add_layer()
    ft.get_children()[-1].add_layer()
    return ft


class TestPdfColumn(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf(orientation='l')

    def test_draw(self):
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            ft = make_ft()
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            ft.graphic.factor = 2
            ft.graphic.draw(self.pdf)

            self.pdf.write(pdf_path)

    def test_add_labels(self):
        with self.file_path(path, 'add_labels', 'pdf') as pdf_path:
            ft = make_ft()
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h', unit=3, show_first_label=True, label_show_interval=5,
                                label_attribute_function=lambda l: setattr(l.font, 'size', 8))
            # self.pdf.draw_ruler('v')
            self.pdf.translate(0, 10)
            ft.graphic.factor = 3
            ft.graphic.add_labels(lambda node: node.fractal_order if node.fractal_order is not None else '',
                                  font_size=8, bottom_margin=1)
            ft.graphic.add_labels(lambda node: round(float(node.value), 2), placement='below', font_size=6,
                                  top_margin=1)
            ft.graphic.change_segment_attributes(bottom_margin=5)

            ft.graphic.draw(self.pdf)
            self.pdf.write(pdf_path)



from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.positioned import RelativeXNotSettableError, RelativeYNotSettableError


class RowContainer(DrawObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._draw_objects = []

    #     relative_x and y are not settable

    @property
    def draw_objects(self):
        return self._draw_objects

    @property
    def relative_x(self):
        return 0

    @relative_x.setter
    def relative_x(self):
        raise RelativeXNotSettableError()

    @property
    def relative_y(self):
        return 0

    @relative_x.setter
    def relative_y(self):
        raise RelativeYNotSettableError()

    def get_relative_x2(self):
        return sum([do.get_width() + do.left_margin + do.right_margin for do in self.draw_objects])

    def get_relative_y2(self):
        return max([do.get_relatvie_y2 + do.top_margin + do.bottom_margin for do in self.draw_objects])

    def add_draw_object(self, draw_object):
        if not isinstance(draw_object, DrawObject):
            raise TypeError()
        self._draw_objects.append(draw_object)


class Canvas(DrawObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._row_containers = []

    @property
    def row_containers(self):
        return self._row_containers

    def add_row_container(self, row_container):
        if not isinstance(row_container, RowContainer):
            raise TypeError()
        self._row_containers.append(row_container)

    def get_relative_x2(self):
        return max([rc.get_width() + rc.left_margin + rc.right_margin for rc in self.row_containers])

    def get_relative_y2(self):
        return sum([rc.get_height + rc.top_margin + rc.bottom_margin for rc in self.row_containers])

    def draw(self, pdf):
        pass



from abc import ABC, abstractmethod

from musurgia.pdf.margined import Margined
from musurgia.pdf.positioned import Positioned


class DrawObject(ABC, Positioned, Margined):
    def __init__(self, *args, **kwargs):
        self._show = True
        super().__init__(*args, **kwargs)

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError(f"show.value must be of type bool not{type(val)}")
        self._show = val

    @abstractmethod
    def get_relative_x2(self):
        raise NotImplementedError()

    @abstractmethod
    def get_relative_y2(self):
        raise NotImplementedError()

    def get_height(self):
        return self.top_margin + self.get_relative_y2() - self.relative_y + self.bottom_margin

    def get_width(self):
        return self.left_margin + self.get_relative_x2() - self.relative_x + self.right_margin

    @abstractmethod
    def draw(self, pdf):
        raise NotImplementedError()

    def get_relative_position(self):
        return {'relative_x': self.relative_x, 'relative_y': self.relative_y}

    def get_margins(self):
        return {'left_margin': self.left_margin, 'top_margin': self.top_margin, 'right_margin': self.right_margin,
                'bottom_margin': self.bottom_margin}





from unittest import TestCase

from quicktions import Fraction

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(value=10)

    def test_1(self):
        self.assertEqual([self.ft], self.ft.get_leaves())

    def test_2(self):
        self.assertEqual([self.ft.value], self.ft.get_leaves(key=lambda leaf: leaf.value))

    def test_3(self):
        self.ft.add_layer()
        [child for child in self.ft.get_children() if child.fractal_order == 1][0].fertile = False
        self.ft.add_layer()
        result = [[Fraction(5, 6), Fraction(5, 3), Fraction(5, 2)], Fraction(5, 3),
                  [Fraction(10, 9), Fraction(5, 3), Fraction(5, 9)]]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.value))



from pathlib import Path

from musurgia.pdf.line import HorizontalLine, SegmentedHorizontalLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.scene import Scene
from musurgia.pdf.segmentedline import SegmentedLine, LineSegment
from musurgia.unittest import TestCase, create_test_path

path = Path(__file__)


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_draw(self):
        pdf_path = create_test_path(path, 'draw.pdf')
        scene = Scene(relative_x=10, relative_y=10)
        scene.add_draw_object(SegmentedHorizontalLine(name='I', lengths=[3, 10, 5, 15]))
        scene.add_draw_object(SegmentedHorizontalLine(name='II', lengths=[3, 10, 5, 15, 20]))
        scene.draw(self.pdf)
        self.pdf.write(pdf_path)
        # self.assertCompareFiles(pdf_path)

    def test_draw_with_margins(self):
        self.assertFalse(True)

    def test_get_height(self):
        scene = Scene()
        do1 = scene.add_draw_object(HorizontalLine(relative_x=10, length=20, top_margin=10, bottom_margin=10))
        do2 = scene.add_draw_object(HorizontalLine(length=10, relative_x=10))
        expected = sum([do.get_height() + do.top_margin + do.bottom_margin + do.relative_y for do in [do1, do2]])
        actual = scene.get_height()
        self.assertEqual(expected, actual)

    def test_get_width(self):
        self.assertFalse(True)

    def test_get_relative_y2(self):
        self.assertFalse(True)

    def test_frame(self):
        self.assertFalse(True)




class Clock(object):
    def __init__(self, duration=None, clock=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._duration = None
        self._clock = None
        self.duration = duration
        self.clock = clock

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, val):
        if val is None:
            val = 0
        self._duration = val
        self._calculate_clock()

    @property
    def clock(self):
        return self._clock

    @clock.setter
    def clock(self, val):
        if val is None:
            val = (0, 0, 0)
        self._clock = val
        self._calculate_duration()

    def _calculate_clock(self):
        h = int(self.duration / 3600.0)
        s = self.duration - h * 3600
        m = int(s / 60.)
        self._clock = (h, m, s)

    def _calculate_duration(self):
        self._duration = self.clock[0] * 3600 + self.clock[1] * 60 + self.clock[2]

    def add(self, clock2):
        output = Clock()
        output.duration = self.duration + clock2.duration
        return output

    def subtract(self, clock2):
        output = Clock()
        output.duration = self.duration - clock2.duration
        return output


class Timing(object):

    @staticmethod
    def get_clock(time, mode='hms'):
        h = int(time / 3600.0)
        s = time - h * 3600
        m = int(s / 60.)
        s = round(s - m * 60, 1)

        if m / 10 == 0 and mode != 'msreduced':
            m = '0' + str(m) + '\''
        else:
            m = str(m) + '\''

        if int(s / 10) == 0 and mode != 'msreduced':
            s = '0' + str(s) + '\"'
        else:
            s = str(s) + '\"'
        h = str(h)

        if not mode or mode == 'hms':
            return h + ':' + m + ':' + s

        elif mode == 'ms':
            return m + ':' + s
        elif mode == 'msreduced':
            if m == '0\'':
                return s
            else:
                return m + ':' + s
        else:
            raise ValueError('mode not known')

    def __init__(self, quarter_duration=None, duration=None, tempo=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_duration = None
        self._duration = None
        self._tempo = None

        self.tempo = tempo
        self.quarter_duration = quarter_duration
        self.duration = duration

    @property
    def duration(self):
        if self._duration is None:
            return self.quarter_duration * 60. / self.tempo

        return self._duration

    @duration.setter
    def duration(self, value):
        if value:
            if self._quarter_duration is not None and self._tempo is not None:
                raise Exception('quarter_duration and tempo are already set. duration cannot be set anymore.')
            self._duration = value

    @property
    def quarter_duration(self):
        if self._quarter_duration is None:
            return self.duration * self.tempo / 60
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, value):
        if value:
            if self._tempo and self._duration:
                raise Exception('duration and tempo are already set. quarter_duration cannot be set anymore.')
            self._quarter_duration = value

    @property
    def tempo(self):
        if self._tempo is None:
            return self.quarter_duration * 60. / self.duration
        return self._tempo

    @tempo.setter
    def tempo(self, value):
        if value:
            if self._quarter_duration and self._duration:
                raise Exception('duration and quarter_duration are already set. tempo cannot be set anymore.')
            self._tempo = value

    @property
    def clock(self):
        h = int(self.duration / 3600.0)
        s = self.duration - h * 3600
        m = int(s / 60.)
        s = round(s - m * 60, 1)
        if m / 10 == 0:
            m = '0' + str(m) + '\''
        else:
            m = str(m) + '\''

        if int(s / 10) == 0:
            s = '0' + str(s) + '\"'
        else:
            s = str(s) + '\"'
        h = str(h)
        return h + ':' + m + ':' + s



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(4, 2, 7, 1, 6, 3, 5))

    def test_1(self):
        actual = self.ft._get_merge_lengths(number_of_children=1, merge_index=4)
        expected = [7]
        self.assertEqual(expected, actual)

    def test_2(self):
        actual = self.ft._get_merge_lengths(number_of_children=2, merge_index=0)
        expected = [6, 1]
        self.assertEqual(expected, actual)

    def test_3(self):
        actual = self.ft._get_merge_lengths(number_of_children=2, merge_index=4)
        expected = [4, 3]
        self.assertEqual(expected, actual)

    def test_4(self):
        actual = self.ft._get_merge_lengths(number_of_children=3, merge_index=2)
        expected = [1, 1, 5]
        self.assertEqual(expected, actual)

    def test_5(self):
        actual = self.ft._get_merge_lengths(number_of_children=3, merge_index=3)
        expected = [2, 1, 4]
        self.assertEqual(expected, actual)

    def test_6(self):
        actual = self.ft._get_merge_lengths(number_of_children=3, merge_index=1)
        expected = [1, 5, 1]
        self.assertEqual(expected, actual)

    def test_7(self):
        actual = self.ft._get_merge_lengths(number_of_children=6, merge_index=3)
        expected = [1, 1, 1, 2, 1, 1]
        self.assertEqual(expected, actual)

    def test_8(self):
        actual = self.ft._get_merge_lengths(number_of_children=7, merge_index=3)
        expected = [1, 1, 1, 1, 1, 1, 1]
        self.assertEqual(expected, actual)

    def test_9(self):
        actual = self.ft._get_merge_lengths(number_of_children=4, merge_index=5)
        expected = [3, 1, 1, 2]
        self.assertEqual(expected, actual)



import os

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(duration=10)

    def test_1(self):
        self.fm.tempo = 70
        self.fm.add_layer()
        self.fm.change_quarter_duration(round(self.fm.quarter_duration))

        xml_path = path + '_test_1.xml'
        score = self.fm.get_score(show_fractal_orders=True)
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        self.fm.tempo = 70
        self.fm.add_layer()
        self.fm.add_layer()
        self.fm.quantize_children()
        xml_path = path + '_test_2.xml'
        score = self.fm.get_children_score(show_fractal_orders=True)
        score.write(xml_path)
        self.assertCompareFiles(xml_path)



from fpdf import FPDF
from fpdf.php import sprintf

from musurgia.pdf.line import HorizontalSegmentedLine, VerticalSegmentedLine
from musurgia.pdf.pdfunit import PdfUnit
from musurgia.pdf.text import PageText, TextLabel


class PageNumber(PageText):
    def __init__(self, value='none', v_position='center', h_position='bottom', *args, **kwargs):
        super().__init__(value=value, v_position=v_position, h_position=h_position, *args, **kwargs)

    def __call__(self, val):
        self.text = val
        self.page = val


class SavedState:
    def __init__(self, pdf):
        self.pdf = pdf

    def __enter__(self):
        self.pdf._push_state()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pdf._pop_state()


class PrepareDrawObject:
    def __init__(self, pdf, draw_object):
        self.pdf = pdf
        self.draw_object = draw_object

    def __enter__(self):
        self.pdf._push_state()
        self.pdf.translate(self.draw_object.relative_x, self.draw_object.relative_y)
        self.pdf.translate(self.draw_object.left_margin, self.draw_object.top_margin)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pdf._pop_state()


class Pdf(FPDF):

    def __init__(self, r_margin=10, t_margin=10, l_margin=10, b_margin=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.page_number = PageNumber('')
        self.r_margin = r_margin
        self.t_margin = t_margin
        self.l_margin = l_margin
        self.b_margin = b_margin
        self.add_page()
        self.show_page_number = False

        self.set_font("Helvetica", "", 10)

    def _pop_state(self):
        self._out(sprintf('Q\n'))

    def _push_state(self):
        self._out(sprintf('q\n'))

    @property
    def show_page_number(self):
        return self._show_page_number

    @show_page_number.setter
    def show_page_number(self, val):
        if not isinstance(val, bool):
            raise TypeError(f"show_page_number.value must be of type bool not{type(val)}")
        self._show_page_number = val

    @property
    def k(self):
        return PdfUnit.get_k()

    @k.setter
    def k(self, val):
        pass

    def add_space(self, val):
        self.y += val

    def add_page(self):
        super().add_page(orientation=self.cur_orientation)

    def prepare_draw_object(self, draw_object):
        pdo = PrepareDrawObject(self, draw_object=draw_object)
        return pdo

    def reset_font(self):
        self._out(sprintf('BT /F%d %.2f Tf ET',
                          self.current_font['i'],
                          self.font_size_pt))

    def clip_rect(self, x, y, w, h):
        x, y, w, h = x * self.k, y * self.k, w * self.k, h * self.k
        self._out(sprintf('%.2f %.2f %.2f %.2f re W n',
                          x * self.k, (self.h - y) * self.k,
                          w * self.k, -h * self.k))

    def draw_page_number(self):
        for page in self.pages:
            self.page = page
            self.page_number(page)
            self.page_number.draw(self)

    def saved_state(self):
        ss = SavedState(self)
        return ss

    def translate(self, dx, dy):
        dx, dy = dx * self.k, dy * self.k
        self._out(sprintf('1.0 0.0 0.0 1.0 %.2F %.2F cm',
                          dx, -dy))

    def translate_page_margins(self):
        self.translate(self.l_margin, self.t_margin)

    def draw_ruler(self, mode='h', unit=10, first_label=0, show_first_label=False, label_show_interval=1,
                   label_attribute_function=None):
        if mode in ['h', 'horizontal']:
            number_of_units = (self.w - self.l_margin - self.r_margin) / unit
        elif mode in ['v', 'vertical']:
            number_of_units = (self.h - self.t_margin - self.b_margin) / unit
        else:
            raise AttributeError()

        partial_segment_length = number_of_units - int(number_of_units)
        lengths = int(number_of_units) * [unit]
        if partial_segment_length:
            lengths += [partial_segment_length * unit]
        if mode in ['h', 'horizontal']:
            ruler = HorizontalSegmentedLine(lengths)
        else:
            ruler = VerticalSegmentedLine(lengths)

        if partial_segment_length:
            ruler.segments[-1].end_mark_line.show = False
        for index, segment in enumerate(ruler.segments):
            if not show_first_label and index == 0:
                pass
            else:
                if index % label_show_interval == 0:
                    tl = TextLabel(index + first_label)
                    if mode in ['v', 'vertical']:
                        tl.placement = 'left'
                        tl.right_margin = 1
                    else:
                        tl.bottom_margin = 1
                    if label_attribute_function:
                        label_attribute_function(tl)
                    segment.start_mark_line.add_text_label(tl)

        with self.saved_state():
            ruler.draw(self)

    def write(self, path):
        if self.show_page_number:
            self.draw_page_number()
        self.output(path, 'F')



from musicscore.basic_functions import lcm, xToD


class Position:
    def __init__(self, value, wheel):
        self._value = None
        self._wheel = None
        self.value = value
        self.wheel = wheel

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val

    @property
    def wheel(self):
        return self._wheel

    @wheel.setter
    def wheel(self, val):
        if not isinstance(val, Wheel):
            raise TypeError('wheel.value must be of type Wheel not{}'.format(type(val)))
        self._wheel = val


class Wheel:
    def __init__(self, size, start=0):
        self._size = None
        self._start = None
        self._number_of_cycles = None
        self.size = size
        self.start = start
        self.index = None

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        if not isinstance(val, int):
            raise TypeError(f'size.value must be of type int not{type(val)}')
        self._size = val

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, val):
        self._start = val

    def get_positions(self, number_of_cycles=None):
        if not number_of_cycles:
            number_of_cycles = self._number_of_cycles
        if not number_of_cycles or not isinstance(number_of_cycles, int) or number_of_cycles < 0:
            raise AttributeError(f'number_of_cycles must be a positive non zero int not {number_of_cycles}')
        position_values = list(range(self.start, number_of_cycles * self.size + self.start, self.size))
        return [Position(value, self) for value in position_values]

    def get_position_values(self):
        position_values = [p.value for p in self.get_positions()]
        return list(dict.fromkeys(position_values))


class GearWheels:
    def __init__(self, wheels):
        self._wheels = None
        self._positions = None
        self._iterator = None

        self.wheels = wheels

    def _set_positions(self):
        _positions = []
        _lcm = lcm(self.gear_sizes)
        for wheel in self.wheels:
            number_of_cycles = (_lcm // wheel.size) + 1
            wheel._number_of_cycles = number_of_cycles
            _positions.extend(wheel.get_positions())
        self._positions = sorted(_positions, key=lambda p: (p.value, p.wheel.index))

    # public properties

    @property
    def gear_sizes(self):
        return [wheel.size for wheel in self.wheels]

    @property
    def wheels(self):
        return self._wheels

    @wheels.setter
    def wheels(self, vals):
        for val in vals:
            if not isinstance(val, Wheel):
                raise TypeError(f'{val} must be of type Wheel not {type(val)} ')
        self._wheels = []
        for index, wheel in enumerate(vals):
            wheel.index = index
            self._wheels.append(wheel)
        self._set_positions()

    # get
    def get_positions(self):
        return self._positions

    def get_position_values(self):
        position_values = [p.value for p in self.get_positions()]
        return list(dict.fromkeys(position_values))

    def get_rhythm(self):
        return xToD(self.get_position_values())



class MarginedOrbjectError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class MarginNotSettableError(MarginedOrbjectError):
    def __init__(self, *args):
        super().__init__(*args)


class Margined:
    def __init__(self, top_margin=None, bottom_margin=None, left_margin=None, right_margin=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._top_margin = None
        self._left_margin = None
        self._bottom_margin = None
        self._right_margin = None

        self.top_margin = top_margin
        self.left_margin = left_margin
        self.bottom_margin = bottom_margin
        self.right_margin = right_margin


    @property
    def top_margin(self):
        if self._top_margin is None:
            return 0
        return self._top_margin

    @top_margin.setter
    def top_margin(self, val):
        self._top_margin = val

    @property
    def left_margin(self):
        if self._left_margin is None:
            return 0
        return self._left_margin

    @left_margin.setter
    def left_margin(self, val):
        self._left_margin = val

    @property
    def bottom_margin(self):
        if self._bottom_margin is None:
            return 0
        return self._bottom_margin

    @bottom_margin.setter
    def bottom_margin(self, val):
        self._bottom_margin = val

    @property
    def right_margin(self):
        if self._right_margin is None:
            return 0
        return self._right_margin

    @right_margin.setter
    def right_margin(self, val):
        self._right_margin = val



import os

from musicscore.musictree.treeinstruments import Violin, Viola, Cello, Percussion, Accordion

from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import PageText
from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusicsquare import Square
from musurgia.timeline.scoretimeline import ScoreTimeLine, ModuleTimeLine

path = str(os.path.abspath(__file__).split('.')[0])

square = Square(duration=100, tree_permutation_order=[3, 1, 2], proportions=[1, 2, 3])

square.change_module_duration(1, 1, 4)
square.get_row(1).set_tempo(50)
square.get_row(2).set_tempo(80)
square.get_row(3).set_tempo(100)

for module in square.modules.values():
    module.quarter_duration = round(module.quarter_duration)

# square.get_row(1).set_score_tempo(50)
# square.get_row(1).set_module_tempo(72)
# square.get_row(2).set_score_tempo(80)
# square.get_row(2).set_module_tempo(72)
# square.get_row(3).set_score_tempo(100)
# square.get_row(3).set_module_tempo(72)


VIOLIN = Violin()
VIOLA = Viola()
CELLO = Cello()
PERCUSSION = Percussion()
ACCORDION = Accordion()


class Test(TestCase):

    def test_1(self):
        pdf = Pdf(orientation='landscape', t_margin=25, l_margin=20)
        title = PageText(value='t e s t', v_position='center', h_position='top', relative_y=-14, font_size=18,
                         font_weight='bold')
        composer = PageText(value='A. G. 2019', v_position='right', h_position='top', relative_y=-14)

        pdf_path = path + '_test_1.pdf'
        stl = ScoreTimeLine(instruments=[VIOLIN, VIOLA, CELLO, PERCUSSION, ACCORDION], units_per_line=60)
        stl.add_module_time_line(
            ModuleTimeLine(start_time=0, module=square.get_module(1, 2), instruments=[VIOLIN, VIOLA, CELLO]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=16, module=square.get_module(2, 3), instruments=[PERCUSSION]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=20, module=square.get_module(3, 1), instruments=[ACCORDION]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=60, module=square.get_module(3, 3),
                           instruments=[VIOLIN, VIOLA, CELLO, PERCUSSION, ACCORDION]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=120, module=square.get_module(1, 3),
                           instruments=[VIOLIN, VIOLA, CELLO]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=135, module=square.get_module(3, 3),
                           instruments=[ACCORDION, PERCUSSION]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=160, module=square.get_module(3, 3),
                           instruments=[ACCORDION, PERCUSSION]))

        stl.draw(pdf)
        for page in pdf.pages:
            pdf.page = page
            title.draw(pdf)
            composer.draw(pdf)
        pdf.write(pdf_path)
        self.assertCompareFiles(actual_file_path=pdf_path)



from musurgia.pdf.drawobject import DrawObject
from musurgia.pdf.segmentedline import LineSegment


class LineGroup(DrawObject):
    def __init__(self, inner_distance=None, bottom_distance=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lines = []
        self._inner_distance = None
        self.inner_distance = inner_distance
        self._bottom_distance = None
        self.bottom_distance = bottom_distance

    def set_distances(self):
        self.set_inner_distance()
        self.set_bottom_distance()

    def set_bottom_distance(self):
        if self.bottom_distance is not None and self.lines:
            self.lines[-1].bottom_margin = self.bottom_distance

    def set_inner_distance(self):
        if self.inner_distance is not None:
            for index, line in enumerate(self.lines):
                line.relative_y = self.relative_y + (index * self.inner_distance)

    def add_line(self, line):
        if not isinstance(line, LineSegment):
            raise TypeError()
        if self.length:
            if line.length != self.length:
                raise ValueError('line.length {} must be {}'.format(line.length, self.length))
        line.relative_y = self.relative_y
        line.relative_x = self.relative_x
        self._lines.append(line)
        self.set_distances()

    @DrawObject.relative_x.setter
    def relative_x(self, val):
        self._relative_x = val
        try:
            for line in self.lines:
                line.relative_x = self.relative_x
        except AttributeError:
            pass

    @DrawObject.relative_y.setter
    def relative_y(self, val):
        self._relative_y = val
        try:
            for line in self.lines:
                line.relative_y = self.relative_y
            self.set_distances()
        except AttributeError:
            pass

    @property
    def lines(self):
        return self._lines

    @property
    def length(self):
        try:
            return self.lines[0].length
        except IndexError:
            return None

    @property
    def inner_distance(self):
        return self._inner_distance

    @inner_distance.setter
    def inner_distance(self, val):
        self._inner_distance = val
        self.set_inner_distance()

    @property
    def bottom_distance(self):
        return self._bottom_distance

    @bottom_distance.setter
    def bottom_distance(self, val):
        self._bottom_distance = val
        self.set_bottom_distance()

    def get_relative_x2(self):
        try:
            return self.lines[0].get_relative_x2()
        except IndexError:
            return None

    def get_relative_y2(self):
        try:
            return self.lines[-1].get_relative_y2() + self.bottom_distance
        except IndexError:
            return None

    def draw(self, pdf):
        old_pdf_x = pdf.x
        for line in self.lines:
            pdf.x = old_pdf_x
            line.draw(pdf)



from unittest import TestCase

from musurgia.tree import Tree


class Test(TestCase):
    def test_1(self):
        t = Tree()
        self.assertIsNone(t.next_leaf)

    def test_2(self):
        t = Tree()
        t.add_child(Tree())
        t.add_child(Tree())
        self.assertIsNone(t.get_children()[1].next_leaf)

    def test_3(self):
        t = Tree()
        t.add_child(Tree())
        t.add_child(Tree())
        self.assertEqual(t.get_children()[0].next_leaf, t.get_children()[1])

    def test_4(self):
        t = Tree()
        t.add_child(Tree())
        t.add_child(Tree())
        t.get_children()[1].add_child(Tree())
        self.assertEqual(t.get_children()[0].next_leaf, t.get_children()[1].get_children()[0])



from musurgia.fractaltree.fractalmusic import FractalMusic

# fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
#                   quarter_duration=30, tempo=80)
fm = FractalMusic(quarter_duration=10, tempo=60)
fm.midi_generator.midi_range = (60, 72)
fm.add_layer()
#
# expected = [fractal_tree.children_generated_midis for fractal_tree in fm.get_children()]
# actual = [fractal_tree.children_generated_midis for fractal_tree in fm.get_children()]
children_midis_1 = [child.children_generated_midis for child in fm.get_children()]
choral_midis_1 = [child.get_choral_midis() for child in fm.get_children()]
choral_midis_2 = [child.get_choral_midis() for child in fm.get_children()]
children_midis_2 = [child.children_generated_midis for child in fm.get_children()]


print(choral_midis_1)
print(children_midis_1)
print(children_midis_2)
# print(choral_midis_2)



import os

from musurgia.unittest import TestCase


path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1_1(self):
        with self.assertRaises(ValueError):
            wrong_path = path + '_test_1'
            self.assertCompareFiles(wrong_path)

    def test_1_2(self):
        actual_file_path = path + '_test_1.txt'
        self.assertCompareFiles(actual_file_path)

    def test_2_1(self):
        actual_file_path = path + '_test_2.txt'
        with self.assertRaises(AssertionError):
            self.assertCompareFiles(actual_file_path)

    def test_2_2(self):
        actual_file_path = path + '_test_2.txt'
        expected_file_path= path + '_test_1_expected.txt'
        self.assertCompareFiles(actual_file_path=actual_file_path, expected_file_path=expected_file_path)



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.fractaltree.fractalmusicsquare import Square

path = str(os.path.abspath(__file__).split('.')[0])


def set_tempo(square):
    for module in square.get_all_modules():
        module.tempo = 72


def add_info(module):
    for index, leaf in enumerate(module.traverse_leaves()):
        leaf.chord.add_lyric(leaf.fractal_order)
        if index == 0:
            leaf.chord.add_words(module.multi)


def generate_score(modules):
    score = TreeScoreTimewise()
    for index, module in enumerate(modules):
        module.get_simple_format(
            layer=module.number_of_layers
        ).to_stream_voice().add_to_score(score=score, part_number=index + 1)

    score.finish()
    partwise = score.to_partwise()
    return partwise


def forward_multi(module, multi_addition):
    m1, m2 = module.multi
    module.multi = (m1, m2 + multi_addition)


class Test(TestCase):
    def setUp(self) -> None:
        self.square = Square(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2), duration=40)
        set_tempo(self.square)
        self.copied_square = self.square.__deepcopy__()

    def test_1(self):
        modules = [self.square.get_module(2, 2), self.copied_square.get_module(2, 2)]
        for module in modules:
            module.add_layer()

        for module in modules:
            add_info(module)

        xml_path = path + '_test_1.xml'
        generate_score(modules).write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        modules = [self.square.get_module(2, 2), self.copied_square.get_module(2, 2)]
        multi_additions = [0, 1]
        for module, multi_addition in zip(modules, multi_additions):
            forward_multi(module, multi_addition)
        for module in modules:
            module.add_layer()

        for module in modules:
            add_info(module)

        xml_path = path + '_test_2.xml'
        generate_score(modules).write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        modules = [self.square.get_module(2, 2), self.copied_square.get_module(2, 2)]
        for module in modules:
            module.children_fractal_values
        multi_additions = [0, 1]
        for module, multi_addition in zip(modules, multi_additions):
            forward_multi(module, multi_addition)
        for module in modules:
            module.add_layer()

        for module in modules:
            add_info(module)

        xml_path = path + '_test_3.xml'
        generate_score(modules).write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        fm_1 = FractalMusic(multi=(1, 2), tempo=60, quarter_duration=10)
        fm_2 = FractalMusic(multi=(1, 1), tempo=60, quarter_duration=10)
        fms = [fm_1, fm_2]

        for fm in fms:
            fm.midi_generator.midi_range = [60, 72]

        fm_2.multi = fm_1.multi

        self.assertEqual(fm_2.children_fractal_values, fm_1.children_fractal_values)
        self.assertEqual(fm_2.children_fractal_orders, fm_1.children_fractal_orders)
        self.assertEqual(fm_2.midi_generator.midi_range, fm_1.midi_generator.midi_range)
        self.assertEqual(fm_2.children_generated_midis, fm_1.children_generated_midis)



import setuptools

setuptools.setup(
    name="musurgia",
    version="2.0.beta",
    author="Alex Gorji",
    author_email="aligorji@hotmail.com",
    description="tools for algorithmic composition",
    url="https://github.com/alexgorji/musurgia.git",
    packages=setuptools.find_packages(),
    install_requires=['quicktions',
                      'musicscore',
                      'prettytable',
                      'fpdf2',
                      'diff-pdf-visually',
                      'matplotlib'
                      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)



import os
from unittest import TestCase

from quicktions import Fraction

from musurgia.fractaltree.fractaltree import FractalTree

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def test_1(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2))
        ft.tempo = 60
        ft.value = 10
        ft.add_layer()
        ft.add_layer()
        self.assertEqual([node.fractal_order for node in ft.traverse_leaves()], [1, 2, 3, 3, 1, 2, 2, 3, 1])

    def test_2(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2))
        ft.tempo = 60
        ft.value = 10
        ft.add_layer()
        ft.add_layer()
        self.assertEqual([node.value for node in ft.traverse_leaves()],
                         [Fraction(5, 6), Fraction(5, 3), Fraction(5, 2), Fraction(5, 6), Fraction(5, 18),
                          Fraction(5, 9),
                          Fraction(10, 9), Fraction(5, 3), Fraction(5, 9)])

    def test_3(self):
        ft = FractalTree(value=10, proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2))
        ft.add_layer()
        ft.get_children()[0].add_layer()
        expected = [[0.8333333333333334, 1.6666666666666667, 2.5], 1.6666666666666667, 3.3333333333333335]
        result = ft.get_leaves(key=lambda leaf: float(leaf.value))
        self.assertEqual(expected, result)



from unittest import TestCase

from quicktions import Fraction

from musurgia.interpolation import InterpolationGroup, RandomInterpolation, Interpolation


class Test(TestCase):

    def test_1(self):
        ig = InterpolationGroup()
        # ig.add_section(InterpolationSection(0, 100, 10))
        # ig.add_section(InterpolationSection(100, 50, 20))
        # ig.add_section(InterpolationSection(50, 200, 10))
        ig.add_section(0, 100, 10)
        ig.add_section(100, 50, 20)
        ig.add_section(50, 200, 10)

        test_case = [float(ig.__call__(x)) for x in range(0, 41)]
        expected = [0.0, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 97.5, 95.0, 92.5, 90.0, 87.5,
                    85.0,
                    82.5, 80.0, 77.5, 75.0, 72.5, 70.0, 67.5, 65.0, 62.5, 60.0, 57.5, 55.0, 52.5, 50.0, 65.0, 80.0,
                    95.0,
                    110.0, 125.0, 140.0, 155.0, 170.0, 185.0, 200.0]

        self.assertEqual(expected, test_case)

    def test_2(self):
        ri = RandomInterpolation(start=[1, 2, 3], end=[7, 7, 7], periodicity=1, duration=20, seed=12)
        test_case = [ri.__call__(x) for x in range(0, 20)]
        expected = [2, 3, 2, 1, 3, 2, 3, 7, 3, 7, 3, 7, 3, 7, 3, 7, 3, 7, 7, 7]
        self.assertEqual(expected, test_case)

    def test_3(self):
        ri = RandomInterpolation(start=[60, 62, 66, 68], end=[67, 69, 73, 75], duration=13, seed=10)
        test_case = [ri.__call__(x) for x in range(0, 13)]
        expected = [60, 68, 67, 62, 66, 69, 67, 68, 66, 73, 69, 68, 73]
        self.assertEqual(expected, test_case)

    def test_4(self):
        interpolation = Interpolation(start=0, end=12, duration=12, grid=0.5)
        test_case = [interpolation.__call__(Fraction(x, 3)) for x in range(0, 12 * 3)]
        expected = [0.0, 0.5, 0.5, 1.0, 1.5, 1.5, 2.0, 2.5, 2.5, 3.0, 3.5, 3.5, 4.0, 4.5, 4.5, 5.0, 5.5, 5.5, 6.0, 6.5,
                    6.5, 7.0, 7.5, 7.5, 8.0, 8.5, 8.5, 9.0, 9.5, 9.5, 10.0, 10.5, 10.5, 11.0, 11.5, 11.5]
        self.assertEqual(expected, test_case)



from musurgia.pdf.drawobjectgroup import DrawObjectGroup
from musurgia.timeline.ruler import Ruler
from musurgia.timeline.voice import Voice


class TimeLineError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TimeLine(DrawObjectGroup):
    def __init__(self, length=None, unit=10, inner_distance=10, bottom_margin=20, *args, **kwargs):
        super().__init__(inner_distance=inner_distance, bottom_margin=bottom_margin, *args, **kwargs)
        self._unit = None
        self.unit = unit
        self._length = None
        self.length = length

    @property
    def ruler(self):
        try:
            return self.draw_objects[0]
        except IndexError:
            return None

    @property
    def voices(self):
        try:
            return self.draw_objects[1:]
        except IndexError:
            return []

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        if val is not None:
            if self._length:
                raise TimeLineError('length can only be set by instantiation')
            self._add_draw_object(Ruler(length=val, unit=self.unit))
            self._length = val

    @property
    def unit(self):
        return self._unit

    def update_units(self):
        for voice in self.draw_objects:
            voice.update_unit(self.unit)

    @unit.setter
    def unit(self, val):
        self._unit = val
        self.update_units()

    def add_voice(self, name):
        voice = Voice(length=self.length, unit=self.unit, name=name)
        return self._add_draw_object(voice)

    def add_draw_object(self, draw_object):
        raise TimeLineError('use add_voice instead')

    # @property
    # def line_groups(self):
    #     if self._line_groups is None:
    #         output = []
    #         for index, line_ in enumerate(self.ruler.lines):
    #             gl = DrawObjectGroup(inner_distance=self.inner_distance, bottom_distance=self.bottom_margin)
    #             gl.add_line(line)
    #             for voice in self.voices:
    #                 gl.add_line(voice.line_segments[index])
    #             output.append(gl)
    #         self._line_groups = output
    #     return self._line_groups

    # def get_relative_x2(self):
    #     raise Exception('timeline has no x2 value')
    #
    # def get_relative_y2(self):
    #     return self._segmented_line_group.get_relative_y2()

    # def get_height(self):
    #     return self.get_relative_y2() - self.relative_y
    #
    # def draw(self, pdf):
    #     for voice in self._segmented_line_group:
    #         voice.
    #         if line_group == self.line_groups[0]:
    #             for fractal_tree, line in zip(self.get_children()[1:], line_group.lines[1:]):
    #                 line.name = fractal_tree.name
    #                 if line.name:
    #                     line.name.relative_y = line.relative_y
    #
    #         line_group.draw_with_break(pdf)
    #         new_x = pdf.x
    #
    #         if line_group._line_break:
    #             for fractal_tree, line in zip(self.get_children(), line_group.lines):
    #                 pdf.x = new_x - line_group.length
    #                 line.name = fractal_tree.name
    #
    #                 if line.name:
    #                     line.name.draw(pdf)
    #             pdf.x = new_x



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm_3 = FractalMusic(value=10)
        self.fm_3.tempo = 60

        self.fm_7 = FractalMusic(value=30, proportions=(1, 2, 3, 4, 5, 6, 7),
                                 tree_permutation_order=(3, 1, 6, 5, 7, 4, 2))
        self.fm_7.tempo = 72

    def test_1(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=0)
        score = fm.get_score(show_fractal_orders=True)
        score = fm.get_score()
        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=1)
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_2.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=2)
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_3.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=3)
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_4.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):
        fm = self.fm_3
        with self.assertRaises(ValueError):
            fm.generate_children(number_of_children=4)

    def test_6(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=(1, 1, 1))
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_6.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_7(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=(0, 1, 2))
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_7.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_8(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=(1, 2, (1, 2, 3)))
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_8.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_9(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=((1, 3), 2, (1, (1, 3), 3)))
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_9.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_10(self):
        fm = self.fm_3
        with self.assertRaisesRegex(expected_exception=ValueError, expected_regex="['reduce', 'merge']"):
            fm.generate_children(number_of_children=3, mode='bla')

    def test_11(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=3, mode='merge')
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_11.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_12(self):
        fm = self.fm_3
        fm.generate_children(number_of_children=2, mode='merge', merge_index=0)
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_12.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_13(self):
        fm = self.fm_7.__deepcopy__()
        fm.add_layer()
        score = fm.get_score(show_fractal_orders=True, layer_number=1)
        fm = self.fm_7.__deepcopy__()
        fm.generate_children(number_of_children=4, mode='merge', merge_index=5)
        for leaf in fm.traverse_leaves():
            leaf.chord.add_lyric(leaf.fractal_order)
        sf = fm.get_simple_format()
        sf.to_stream_voice().add_to_score(score, part_number=2)

        xml_path = path + '_test_13.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_14(self):
        fm = self.fm_7.__deepcopy__()
        fm.add_layer()
        score = fm.get_score(show_fractal_orders=True, layer_number=1)
        fm = self.fm_7.__deepcopy__()
        fm.generate_children(number_of_children=(4, 2), mode='merge', merge_index=5)

        for leaf in fm.traverse_leaves():
            leaf.chord.add_lyric(leaf.fractal_order)
        sf = fm.get_simple_format(layer=1)
        sf.to_stream_voice().add_to_score(score, part_number=2)
        sf = fm.get_simple_format(layer=2)
        sf.to_stream_voice().add_to_score(score, part_number=3)

        xml_path = path + '_test_14.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_15(self):
        self.fm_7.quarter_duration = 10
        score = TreeScoreTimewise()
        score.add_title('reduce backwards')
        score.page_style.orientation = 'portrait'
        score.set_time_signatures(barline_style='light-light', quarter_durations=self.fm_7.quarter_duration)
        for i in range(1, 8):
            fm = self.fm_7.__deepcopy__()
            fm.generate_children(number_of_children=i, mode='reduce_backwards')
            for leaf in fm.traverse_leaves():
                leaf.chord.add_lyric(leaf.fractal_order)
            sf = fm.get_simple_format(layer=fm.number_of_layers)
            sf.to_stream_voice().add_to_score(score, part_number=i)

        xml_path = path + '_test_15.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_16(self):
        self.fm_7.quarter_duration = 10
        score = TreeScoreTimewise()
        score.add_title('reduce forwards')
        score.page_style.orientation = 'portrait'
        score.set_time_signatures(barline_style='light-light', quarter_durations=self.fm_7.quarter_duration)
        for i in range(1, 8):
            fm = self.fm_7.__deepcopy__()
            fm.generate_children(number_of_children=i, mode='reduce_forwards')
            for leaf in fm.traverse_leaves():
                leaf.chord.add_lyric(leaf.fractal_order)
            sf = fm.get_simple_format(layer=fm.number_of_layers)
            sf.to_stream_voice().add_to_score(score, part_number=i)

        xml_path = path + '_test_16.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_17(self):
        self.fm_7.quarter_duration = 10
        score = TreeScoreTimewise()
        score.add_title('reduce sieve')
        score.page_style.orientation = 'portrait'
        score.set_time_signatures(barline_style='light-light', quarter_durations=self.fm_7.quarter_duration)
        for i in range(1, 8):
            fm = self.fm_7.__deepcopy__()
            fm.generate_children(number_of_children=i, mode='reduce_sieve')
            for leaf in fm.traverse_leaves():
                leaf.chord.add_lyric(leaf.fractal_order)
            sf = fm.get_simple_format(layer=fm.number_of_layers)
            sf.to_stream_voice().add_to_score(score, part_number=i)

        xml_path = path + '_test_17.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)




from pathlib import Path

from musurgia.pdf.line import HorizontalSegmentedLine, VerticalSegmentedLine
from musurgia.pdf.masterslave import PositionMaster
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import TextLabel
from musurgia.unittest import TestCase

path = Path(__file__)


class DummyPositionMaster(PositionMaster):
    def get_slave_position(self, slave, position):
        if position == 'x':
            return 0
        elif position == 'y':
            return 0


class TestTextLabel(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_draw(self):
        t = TextLabel(master=DummyPositionMaster(), name='t1', text='Fox is going to be dead.')

        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            t.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_multiple(self):
        t1 = TextLabel(master=DummyPositionMaster(), name='t1', text='Fox is going to be dead.')
        t1.top_margin = -5
        t2 = TextLabel(master=DummyPositionMaster(), name='t2', text='What should we do??')
        t3 = TextLabel(master=DummyPositionMaster(), name='t3', text='What should we do??')
        with self.file_path(path, 'draw_multiple', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            t1.draw(self.pdf)
            self.pdf.translate(0, t1.get_height())
            t2.draw(self.pdf)
            self.pdf.translate(0, t2.get_height())
            t3.draw(self.pdf)
            self.pdf.translate(0, t3.get_height())
            self.pdf.write(pdf_path)



from musurgia.pdf.font import Font
from musurgia.pdf.masterslave import PositionSlave
from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.pdfunit import PdfUnit


class Text(DrawObject):
    DEFAULT_FONT_FAMILY = 'Helvetica'
    DEFAULT_FONT_SIZE = 10
    DEFAULT_FONT_WEIGHT = 'medium'
    DEFAULT_FONT_STYLE = 'regular'

    def __init__(self, value, font_family=None, font_weight=None, font_style=None,
                 font_size=None, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.font = Font()
        self.font_family = font_family
        self.font_weight = font_weight
        self.font_style = font_style
        self.font_size = font_size
        self._value = None
        self.value = value

    @property
    def font_family(self):
        return self.font.family

    @font_family.setter
    def font_family(self, val):
        if val is None:
            val = self.DEFAULT_FONT_FAMILY
        self.font.family = val

    @property
    def font_size(self):
        return self.font.size

    @font_size.setter
    def font_size(self, val):
        if val is None:
            val = self.DEFAULT_FONT_SIZE
        self.font.size = val

    @property
    def font_weight(self):
        return self.font.weight

    @font_weight.setter
    def font_weight(self, val):
        if val is None:
            val = self.DEFAULT_FONT_WEIGHT
        self.font.weight = val

    @property
    def font_style(self):
        return self.font.style

    @font_style.setter
    def font_style(self, val):
        if val is None:
            val = self.DEFAULT_FONT_STYLE
        self.font.style = val

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = str(val)

    def get_text_width(self):
        return self.font.get_text_pixel_width(self.value) / PdfUnit.get_k()

    def get_text_height(self):
        return self.font.get_text_pixel_height(self.value) / PdfUnit.get_k()

    def get_relative_x2(self):
        return self.relative_x + self.get_text_width()

    def get_relative_y2(self):
        return self.relative_y + self.get_text_height()

    def draw(self, pdf):
        if pdf.k != PdfUnit.get_k():
            raise AttributeError('wrong pdf.k!')
        if self.show:
            pdf.reset_font()
            style = ""
            if self.font.style == 'italic':
                style += 'I'
            if self.font.weight == 'bold':
                style += 'B'
            pdf.set_font(self.font.family, style=style, size=self.font_size)
            with pdf.prepare_draw_object(self):
                pdf.text(x=0, y=0, txt=self.value)


class TextLabel(PositionSlave, Text):
    def __init__(self, text, placement='above', *args, **kwargs):
        super().__init__(value=text, *args, **kwargs)
        self._placement = None
        self.placement = placement

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, val):
        permitted = ['above', 'below', 'left']
        if val not in permitted:
            raise ValueError(f'placement.value {val} must be in {permitted}')
        self._placement = val


class PageText(Text):
    def __init__(self, value, v_position=None, h_position=None, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)
        self.v_position = v_position
        self.h_position = h_position

    def get_text_physical_length(self):
        return (len(self.text)) * self.font_size / 6

    def draw(self, pdf):
        if self.v_position == 'center':
            pdf.x = (pdf.w / 2) - self.get_text_physical_length() / 2
        elif self.v_position == 'left':
            pdf.x = pdf.l_margin
        elif self.v_position == 'right':
            pdf.x = pdf.w - pdf.r_margin - self.get_width()
        else:
            pass

        if self.h_position == 'top':
            pdf.y = pdf.t_margin
        elif self.h_position == 'bottom':
            pdf.y = pdf.h - pdf.b_margin
        else:
            pass
        super().draw(pdf)
